/*
 * Apollo dn3000/3500 emulator
 *
 * Copyright 2010 Jim Rees jim.rees@rees.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define DISP_WIDTH 2048
#define DISP_VIS_WIDTH 1280
#define DISP_HEIGHT 1024
#define DISP_SIZE (2048 * DISP_HEIGHT / 8)

#define avarg() (ac--, av++, av[0])
#define getav(ac, av) (ac > 1 && av[1][0] == '-' ? avarg() : NULL)

#define is_physmem(a) ((a) >= phys_base && (a) < phys_base + phys_size)
#define is_monomem(a) ((a) >= disp_base && a < disp_base + DISP_SIZE)
#define is_monoctl(a) ((a) >= 0x5d800 && (a) < 0x5dc08)
#define phys_ptr(a) ((a) - phys_base + phys_mem)
#define disp_ptr(a) ((a) - disp_base + disp_mem)

struct rfc_head {
    long st, en, eol, ts;
};

#define ST_OFF		0
#define ST_RESET	1
#define ST_STEP		2
#define ST_RUN		3
#define ST_STOP		4
#define ST_QUIT		5

extern int vflag, runstate, magic, vis_width, sau, quiet, dbgmask;
extern void *disp_mem;
extern unsigned long disp_base;

void emu_msleep(int msec);

void sio_init(int swap_sio, int hook_tty);
void sio_run();
void sio_inskey(unsigned int c);
unsigned int sio_read_memory_8(unsigned int addr);
void sio_write_memory_8(unsigned int addr, unsigned int data);
void sio_end();
extern int key_vflag;

void mono_init();
void mono_run();
unsigned int mono_read_memory_8(unsigned int addr);
void mono_write_memory_8(unsigned int addr, unsigned int data);
void mono_write_memory_16(unsigned int addr, unsigned int data);
void fb_write_memory_8(unsigned int addr, unsigned int data);
void fb_write_memory_16(unsigned int addr, unsigned int data);
void fb_write_memory_32(unsigned int addr, unsigned int data);
void mono_print_screen(int x0, int y0, int x1, int y1);
void mono_end();
extern int mono_vflag, fb_vflag, bltv_x, bltv_y;

int vnc_init(int vncfd);
void vnc_run(int block);
void led_write_memory_8(unsigned int addr, unsigned int data);
void vnc_bell();
void vnc_end();
extern int vnc_vflag;

int disk_init(int dimgfd);
unsigned int disk_read_memory_8(unsigned int addr);
unsigned int disk_read_memory_16(unsigned int addr);
void disk_write_memory_8(unsigned int addr, unsigned int data);
void disk_write_memory_16(unsigned int addr, unsigned int data);

int ct_init();
unsigned int ct_read_memory_8(unsigned int addr);
void ct_write_memory_8(unsigned int addr, unsigned int data);

void pic_write_memory_8(unsigned int addr, unsigned int data);
