/*
 * mono display handler
 *
 * Copyright 2010 Jim Rees jim.rees@rees.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _XOPEN_SOURCE 600

#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>

#include "m68k.h"
#include "dn3000.h"

void *disp_mem;
unsigned long disp_base = 0xfa0000;
int mono_vflag, fb_vflag;
int rshift;
int bltv_x = -1, bltv_y = -1;
unsigned int blt_ctl_addr = 0x5dc00;
unsigned int blt_mask_addr = 0x5d800;
unsigned int misc_addr = 0x5dc02;
unsigned int reg04_addr = 0x5dc04;
unsigned int reg06_addr = 0x5dc06;
unsigned int bltv_addr;
unsigned char blt_ctl_reg, amode, misc_reg, reg04, reg06, fx;
u_int16_t blt_mask;
int xdst, ydst;

int nblt;

void
mono_init()
{
    if (disp_mem == NULL)
	disp_mem = malloc(DISP_SIZE);
    memset(disp_mem, 0, DISP_SIZE);
    bltv_addr = (bltv_y * DISP_WIDTH + bltv_x) / 8;
}

void
mono_run()
{
}

unsigned int
mono_read_memory_8(unsigned int addr)
{
    unsigned int data = 0x0;

    if (addr == 0x5d801) {
	/* Mono display type register */
	data = 0x9;
    } else if (addr == blt_ctl_addr) {
	data = blt_ctl_reg;
    } else if (addr == misc_addr) {
	data = misc_reg;
    } else if (addr == reg04_addr) {
	data = reg04;
    } else if (addr == reg06_addr) {
	data = reg06;
    } else
	printf("mono8 %x\n", addr);

    return data;
}

void
mono_write_memory_8(unsigned int addr, unsigned int data)
{
    if (addr == blt_ctl_addr) {
	if (blt_ctl_reg != data) {
	    blt_ctl_reg = data;
	    amode = data >> 5;
	    rshift = data & 0x1f;
	    if (rshift & 0x10)
		rshift |= 0xe0;
	    if (mono_vflag)
		printf("amode=%x rshift=%d\n", amode, rshift);
	}
    } else if (addr == misc_addr) {
	if (misc_reg != data) {
	    misc_reg = data;
	    ydst = (data & 0x40) << 3;
	    fx = (data & 0xf);
	    if (fx != 0xb)
		printf("*** fx=%x ***\n", fx);
	    nblt = 0;
	}
    } else if (addr == reg04_addr) {
	reg04 = data;
	printf("reg04: %x\n", data);
    } else if (addr == reg06_addr) {
	reg06 = data;
	printf("reg06: %x\n", data);
    } else
	printf("mono8 %x: %x\n", addr, data);
}

void
mono_write_memory_16(unsigned int addr, unsigned int data)
{
    unsigned int hdata = htons(data);

    if (addr == blt_mask_addr) {
	if (mono_vflag && blt_mask != data) {
	    printf("blt_mask=%x\n", data);
	}
	blt_mask = data;
    } else if (mono_vflag)
	printf("mono16 %x: %x\n", addr, hdata);
}

void
fb_write_memory_8(unsigned int addr, unsigned int data)
{
    if (addr - disp_base == bltv_addr)
	printf("WriteDisp8 %x: %x\n", addr, data);

    if (amode == 0x7 && addr == 0xfdc000 && data == 0xfe) {
	/*
	 * I don't know what's going on here.  I think data is a length
	 * field, and we're supposed to fill the line with something to that
	 * length.  For now I'm just going to blank out the last line of
	 * text since the prom only uses this for scrolling.
	 */
	memset(disp_ptr(addr), 0, 256 * 13);
    }

    if (addr - disp_base == bltv_addr)
	printf("m8  amode=%d blt_mask=%04x addr=%x data=%x\n",
	       amode, blt_mask, addr, data);
    nblt++;
}

void
fb_write_memory_16(unsigned int addr, unsigned int data)
{
    unsigned int hdata = htons(data);
    u_int16_t *pdst, srcpx, dstpx, dstbg;
    int rno, xsrc, ysrc;

    rno = addr - disp_base;
    xsrc = (rno % 256) * 8;
    ysrc = rno / 256;

    if (amode == 0x4) {
	/* double access blt */
	xdst = data & 0x7f;
	ydst = (ydst & 0x200) | ((data & 0xff80) >> 7);

	srcpx = ntohs(*((u_int16_t *) disp_ptr(addr) - 1)) << (16 - rshift);
	srcpx |= ntohs(*(u_int16_t *) disp_ptr(addr)) >> rshift;
	srcpx &= ~blt_mask;

	pdst = (u_int16_t *) disp_mem + ydst * 128 + xdst;
	dstpx = ntohs(*pdst);
	dstbg = dstpx & blt_mask;
	dstpx &= ~blt_mask;

	*pdst = htons(dstbg | (srcpx & ~blt_mask));

	if ((xdst == bltv_x * 16 && ydst > 959) || (dbgmask & 1)) {
	    printf("dblt amode=%d mask=%04x ctl=%02x rreg=%02x src=%d,%d dst=%d,%d\n",
		   amode, blt_mask, blt_ctl_reg, misc_reg,
		   xsrc, ysrc, xdst * 16, ydst);
	}
    } else if (amode == 0x7) {
	/* normal access */
	*(u_int16_t *) disp_ptr(addr) = hdata;
    }
    if (amode != 4 && addr - disp_base == bltv_addr)
	printf("m16 amode=%d blt_mask=%04x src=%d,%d data=%x\n",
	       amode, blt_mask, xsrc, ysrc, data);
    nblt++;
}

void
fb_write_memory_32(unsigned int addr, unsigned int data)
{
#ifdef BLT_32
    *(u_int32_t *) disp_ptr(addr) = htonl(data);
#endif
    if (addr - disp_base == bltv_addr)
	printf("m32 amode=%d blt_mask=%x addr=%x data=%x\n",
	       amode, blt_mask, addr, data);
}

void
mono_print_screen(int x0, int y0, int x1, int y1)
{
    unsigned long *lp, d;
    int x, y, lw, vlw;
    FILE *f;

    lw = DISP_WIDTH / (sizeof d * 8);
    vlw = (x1 - x0) / (sizeof d * 8);

    f = popen("pnmtopng >dn3000.png", "w");
    if (f == NULL)
	return;

    fprintf(f, "P4\n%d %d\n", x1 - x0, y1 - y0);
    for (y = y0; y < y1; y++) {
	for (x = x0; x < x0 + vlw; x++) {
	    lp = (unsigned long *) disp_mem + y * lw + x;
	    d = ~(*lp);
	    fwrite(&d, sizeof d, 1, f);
	}
    }

    fclose(f);
}

void
mono_end()
{
    if (disp_mem) {
	free(disp_mem);
	disp_mem = NULL;
    }
}
