/*
 * sio handler
 *
 * Copyright 2010 Jim Rees jim.rees@rees.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _XOPEN_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <termios.h>

#include "m68k.h"
#include "dn3000.h"

struct termios termio0;
int key_vflag, kfd = -1, sion, n_sio_char_read;
char kcin;

void
sio_init(int swap_sio, int hook_tty)
{
    struct termios t1;

    if (hook_tty) {
	kfd = fileno(stdin);
	tcgetattr(kfd, &termio0);
	t1 = termio0;
	t1.c_iflag &= ~(INLCR|IGNCR|ICRNL|IXON);
	t1.c_iflag |= (ISTRIP|IGNBRK|IGNPAR);
	t1.c_oflag |= ONLCR;
	t1.c_lflag &= ~(ECHO|ECHONL|ICANON|IEXTEN);
	tcsetattr(kfd, TCSANOW, &t1);
    }
    if (swap_sio)
	sion = 0x10;
}

void
sio_run()
{
    fd_set fdset;
    struct timeval tv;
    char buf[1];
    int n;

    if (kfd < 0)
	return;

    FD_ZERO(&fdset);
    FD_SET(kfd, &fdset);
    tv.tv_sec = tv.tv_usec = 0;
    if (select(kfd + 1, &fdset, NULL, NULL, &tv) < 1)
	return;

    if (!FD_ISSET(kfd, &fdset))
	return;

    n = read(kfd, buf, sizeof buf);
    if (n == 1) {
	if (buf[0] == 0x1b)
	    runstate = ST_QUIT;
	else
	    sio_inskey(buf[0]);
    }
}

void
sio_inskey(unsigned int c)
{
    if (key_vflag)
	printf("keypress 0x%02x\n", c);

    if (kcin != '\0')
	printf("sio_inskey: overrun\n");
    kcin = c;
}

unsigned int
sio_read_memory_8(unsigned int addr)
{
    unsigned int data = 0x0;
    static unsigned int sio_stat_b = 0x84;  /* break | xmit rdy */

    addr &= 0xff;

    if (addr == (0x13 ^ sion)) {
	/* sio0 status register A; xmit rdy */
	data = 0x4;
    } else if (addr == (0x03 ^ sion)) {
	/* sio1 status register B */
	if (kcin != '\0')
	    sio_stat_b |= 0x1;
	data = sio_stat_b;
	sio_stat_b &= ~0x80;
    } else if (addr == (0x07 ^ sion)) {
	/* tx/rx data */
	if (kcin != '\0') {
	    data = kcin;
	    kcin = '\0';
	}
	sio_stat_b &= ~0x1;
	n_sio_char_read++;
    } else if (addr == 0x1b) {
	/* input port register; cts | dcd a/b */
	data = 0xf;
    } else if (addr == 0x1d)
	data = 0x0;

    return data;
}

void
sio_write_memory_8(unsigned int addr, unsigned int data)
{
    int sion = 0;

    if (addr & 0x100)
	sion += 2;
    if (addr & 0x10)
	sion++;

    if ((addr & 0x7f) == 0x07)
	printf("sio%d: %x\n", sion, data);
}

void
sio_end()
{
    if (kfd >= 0) {
	tcsetattr(kfd, TCSANOW, &termio0);
	printf("n_sio_char_read=%d\n", n_sio_char_read);
	kfd = -1;
    }
}
